import collections.abc
import os
from functools import partial, wraps
from types import ModuleType
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

from .exceptions import MykeNotFoundError, NoTasksFoundError, TaskAlreadyRegisteredError
from .io import read, write
from .sh import sh
from .utils import (
    MykeSourceFileLoader,
    convert_to_command_string,
    make_executable,
    split_and_trim_text,
)

TASKS: Dict[str, Callable[..., None]] = {}

MYKE_VAR_NAME: str = "myke"

ROOT_TASK_KEY: str = "__root__"


def add_tasks(*args: Callable[..., Any], **kwargs: Callable[..., Any]) -> None:

    kwargs.update({x.__name__: x for x in args})

    kwargs = {
        (ROOT_TASK_KEY if k == ROOT_TASK_KEY else convert_to_command_string(k)): v
        for k, v in kwargs.items()
    }

    for k, v in kwargs.items():
        v_existing: Optional[Callable[..., Any]] = TASKS.get(k, None)
        if v_existing:
            raise TaskAlreadyRegisteredError(
                f"Failed to import module '{v.__module__}': "
                f"Command '{k}' already defined from module '{v_existing.__module__}'"
            )
        TASKS[k] = v


def import_module(*mykefiles: str, overwrite: bool = False) -> None:
    n_tasks_before: int = len(TASKS)
    for m in mykefiles:
        if m.startswith("https://"):
            m = install_module(m, overwrite=overwrite)

        loader = MykeSourceFileLoader(os.path.relpath(m), m)
        mod: ModuleType = ModuleType(loader.name)
        loader.exec_module(mod)

        if not hasattr(mod, MYKE_VAR_NAME):
            raise MykeNotFoundError(m)

        n_tasks_after: int = len(TASKS)
        if n_tasks_after <= n_tasks_before:
            raise NoTasksFoundError(m)
        n_tasks_before = n_tasks_after


def install_module(
    url: str,
    path: Optional[str] = None,
    if_not_exists: bool = True,
    overwrite: bool = False,
) -> str:
    if not url.startswith("https://"):
        raise ValueError("Download URLs must start with 'https://'")

    if path is None:
        path = "tasks"

        if not os.path.exists(path):
            os.mkdir(path)

        path = os.path.join(path, os.path.basename(url))

    if not overwrite and os.path.exists(path) and if_not_exists:
        return path

    resp_text: str = read.url(url)

    import_myke: str = f"import {MYKE_VAR_NAME}"
    if import_myke not in resp_text:
        raise MykeNotFoundError(f'"{import_myke}" not found response text of {url}')

    write.text(resp_text, path=path, overwrite=overwrite)

    make_executable(path)

    return path


def task(
    func: Optional[Callable[..., Any]] = None,
    *,
    name: Optional[str] = None,
    root: Optional[bool] = False,
) -> Union[Callable[..., Any], Callable[..., Callable[..., Any]]]:

    if not func:
        return partial(task, name=name, root=root)

    if root:
        name = ROOT_TASK_KEY

    if name:
        add_tasks(**{name: func})
    else:
        add_tasks(func)

    return func


def task_sh(
    func: Optional[Callable[..., Union[str, Sequence[str]]]] = None,
    *,
    name: Optional[str] = None,
    root: Optional[bool] = False,
    capture_output: Optional[bool] = False,
    echo: Optional[bool] = True,
    check: Optional[bool] = True,
    cwd: Optional[str] = None,
    env: Optional[Dict[str, str]] = None,
    env_update: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None,
    executable: Optional[str] = None,
) -> Union[
    Callable[..., Tuple[Optional[str], Optional[str], int]],
    Callable[..., Callable[..., Tuple[Optional[str], Optional[str], int]]],
]:
    if not func:
        return partial(
            task_sh,
            name=name,
            root=root,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    @wraps(func)
    def _inner_func(
        *args: Any, **kwargs: Any
    ) -> Tuple[Optional[str], Optional[str], int]:
        assert func

        func_script: Union[str, Sequence[str]] = func(*args, **kwargs)

        if not isinstance(func_script, str) and not isinstance(
            func_script, collections.abc.Sequence
        ):
            raise TypeError(
                "Expected a string or list of strings to be returned for `sh` function"
            )

        return sh(
            func_script,
            capture_output=capture_output,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    if not name:
        name = func.__name__

    return task(_inner_func, name=name, root=root)


def _task_sh_stdout_lines(
    func: Optional[Callable[..., Union[str, Sequence[str]]]] = None,
    *,
    name: Optional[str] = None,
    root: Optional[bool] = False,
    echo: Optional[bool] = False,
    check: Optional[bool] = True,
    cwd: Optional[str] = None,
    env: Optional[Dict[str, str]] = None,
    env_update: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None,
    executable: Optional[str] = None,
    join_lines: Optional[bool] = False,
) -> Union[
    Callable[..., Union[str, List[str]]],
    Callable[..., Callable[..., Union[str, List[str]]]],
]:
    if not func:
        return partial(
            task_sh_stdout_lines,
            name=name,
            root=root,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

    @wraps(func)
    def _inner_func(*args: Any, **kwargs: Any) -> Union[str, List[str]]:
        assert func
        stdout, *_ = sh(
            func(*args, **kwargs),
            capture_output=True,
            echo=echo,
            check=check,
            cwd=cwd,
            env=env,
            env_update=env_update,
            timeout=timeout,
            executable=executable,
        )

        stdout_split: List[str] = split_and_trim_text(stdout)

        return os.linesep.join(stdout_split) if join_lines else stdout_split

    if not name:
        name = func.__name__

    return task(_inner_func, name=name, root=root)


@wraps(_task_sh_stdout_lines)
def task_sh_stdout_lines(
    *args: Any, **kwargs: Any
) -> Union[Callable[..., List[str]], Callable[..., Callable[..., List[str]]]]:
    return _task_sh_stdout_lines(*args, **kwargs)  # type: ignore


@wraps(_task_sh_stdout_lines)
def task_sh_stdout(
    *args: Any, **kwargs: Any
) -> Union[Callable[..., str], Callable[..., Callable[..., str]]]:
    kwargs["join_lines"] = True
    return _task_sh_stdout_lines(*args, **kwargs)  # type: ignore



import collections.abc
import os
from functools import wraps
from typing import Any, Dict, Iterable, Mapping, Optional, Sequence, Union


def text(txt: Optional[str] = None, **kwargs: Any) -> None:
    args = [txt] if txt else []
    print(*args, **kwargs)


@wraps(text)
def _print(txt: str, print_kwargs: Optional[Dict[str, Any]] = None) -> None:
    if not print_kwargs:
        print_kwargs = {}
    print(txt, **print_kwargs)


def lines(
    seq: Sequence[str],
    linesep: str = os.linesep,
    print_kwargs: Optional[Dict[str, Any]] = None,
) -> None:
    _print(linesep.join(seq), print_kwargs)


def json(
    obj: Any, print_kwargs: Optional[Dict[str, Any]] = None, **kwargs: Any
) -> None:
    import json as _json

    _print(_json.dumps(obj, **kwargs), print_kwargs)


def table(
    obj: Union[Iterable[Iterable[Any]], Mapping[str, Iterable[Any]]],
    print_kwargs: Optional[Dict[str, Any]] = None,
    **kwargs: Any
) -> None:
    from tabulate import tabulate

    if (
        obj
        and isinstance(obj, collections.abc.Sequence)
        and isinstance(obj[0], dict)
        and "headers" not in kwargs
    ):
        kwargs["headers"] = "keys"

    _print(tabulate(obj, **kwargs), print_kwargs)


def pretty(
    obj: Any, print_kwargs: Optional[Dict[str, Any]] = None, **kwargs: Any
) -> None:
    from pprint import pformat

    _print(pformat(obj, **kwargs), print_kwargs)



import os
from typing import Any, List, Optional, Union

from ..io import read
from ..main import DEFAULT_MYKEFILE
from ..utils import make_executable


def text(
    content: Union[str, bytes],
    path: str,
    append: bool = False,
    overwrite: bool = False,
    **kwargs: Any
) -> None:
    mode_default: str = "w"

    mode: str = kwargs.pop("mode", mode_default)

    if os.path.exists(path):
        if overwrite:
            os.remove(path)
        elif append and mode == mode_default:
            mode = "a"
        else:
            raise FileExistsError(path)

    if mode == mode_default:
        if isinstance(content, bytes):
            mode += "b"

        if not os.path.exists(path):
            mode += "+"

    with open(path, mode=mode, **kwargs) as f:
        f.write(content)


def lines(
    content: List[str],
    path: str,
    append: bool = False,
    overwrite: bool = False,
    **kwargs: Any
) -> None:
    if isinstance(content, str):
        raise TypeError("expected a list of strings, not a string")
    text(
        path=path,
        content=os.linesep.join(content),
        append=append,
        overwrite=overwrite,
        **kwargs
    )


def mykefile(path: Optional[str] = None) -> None:
    if not path:
        path = os.getcwd()

    if os.path.isdir(path):
        path = os.path.join(path, DEFAULT_MYKEFILE)

    text(
        path=path,
        content="""#!/usr/bin/env python3
# type: ignore

import myke


@myke.task(root=True)
def setup() -> None:
    print("charging lazer beamz")


@myke.task
def print_hello(
    name = myke.arg(default="world", env_var="HELLO_NAME", flags=["-n", "--name"])
):
    print(f"hello {name}")

@myke.task_sh
def echo_hello(
    name = myke.arg(default="world", env_var="HELLO_NAME", flags=["-n", "--name"])
):
    return f"echo 'hello {name}'"

if __name__ == "__main__":
    myke.main(__file__)
""",
    )

    make_executable(path)


def url(
    addr: str,
    path: Optional[str] = None,
    append: bool = False,
    overwrite: bool = False,
    **kwargs: Any
) -> str:
    if not path:
        path = os.getcwd()

    if os.path.isdir(path):
        path = os.path.join(path, os.path.basename(addr))

    resp: str = read.url(addr)

    text(content=resp, path=path, append=append, overwrite=overwrite, **kwargs)

    return path



__all__ = ["CalledProcessError"]

import os
import subprocess


class MykeNotFoundError(Exception):
    ...


class TaskAlreadyRegisteredError(Exception):
    ...


class NoTasksFoundError(Exception):
    ...


class CalledProcessError(subprocess.CalledProcessError):
    def __str__(self) -> str:
        if self.returncode and self.returncode < 0:
            return super().__str__()
        else:
            msg: str = (
                f"Command '{self.cmd}' returned non-zero exit status {self.returncode}:"
            )
            for attr in "stdout", "stderr":
                attr_value: str = getattr(self, attr)
                if attr_value:
                    attr_value.rstrip(os.linesep)
                else:
                    attr_value = ""
                msg += f"{os.linesep+attr}:{os.linesep+attr_value}"
            return msg



__all__ = [
    "__version__",
    "TASKS",
    "task",
    "task_sh",
    "task_sh_stdout",
    "task_sh_stdout_lines",
    "add_tasks",
    "import_module",
    "install_module",
    "sh",
    "sh_stdout",
    "sh_stdout_lines",
    "arg",
    "main",
    "read",
    "write",
    "utils",
    "exceptions",
    "cache",
    "print",
]

from functools import lru_cache as cache

from yapx import arg

from . import exceptions, utils
from .io import print, read, write
from .main import __version__, main
from .sh import sh, sh_stdout, sh_stdout_lines
from .tasks import (
    TASKS,
    add_tasks,
    import_module,
    install_module,
    task,
    task_sh,
    task_sh_stdout,
    task_sh_stdout_lines,
)



__all__ = ["__version__", "main", "sys"]

import os
import sys
from dataclasses import dataclass
from typing import Any, Callable, List, Optional

import yapx

from .exceptions import NoTasksFoundError
from .io import print, read, write
from .tasks import MYKE_VAR_NAME, ROOT_TASK_KEY, TASKS, import_module

DEFAULT_MYKEFILE: str = "Mykefile"

with open(
    os.path.join(os.path.dirname(__file__), "VERSION"), mode="r", encoding="utf-8"
) as _f:
    __version__: str = _f.read().strip()


def main(_file: Optional[str] = None) -> None:
    @dataclass
    class MykeArgs(yapx.types.Dataclass):
        file: str = yapx.arg(
            default=_file if _file else DEFAULT_MYKEFILE,
            flags=["--myke-file"],
            env_var="MYKE_FILE",
            group="myke args",
        )
        file_paths: List[str] = yapx.arg(
            default=lambda: [os.path.expanduser("~"), os.getcwd()],
            flags=["--myke-file-paths"],
            env_var="MYKE_FILE_PATHS",
            group="myke args",
        )
        env_file: str = yapx.arg(
            default=None,
            flags=["--myke-env-file"],
            env_var="MYKE_ENV_FILE",
            group="myke args",
        )
        no_pydantic: bool = yapx.arg(
            default=False, group="myke args", flags=["--myke-no-pydantic"]
        )
        task_help: bool = yapx.arg(
            default=False, group="myke args", exclusive=True, flags=["-h", "--help"]
        )
        task_help_all: bool = yapx.arg(
            default=False, group="myke args", exclusive=True, flags=["--help-all"]
        )
        help: bool = yapx.arg(
            default=False, group="myke args", exclusive=True, flags=["--myke-help"]
        )
        version: bool = yapx.arg(
            default=False, group="myke args", exclusive=True, flags=["--myke-version"]
        )
        create: bool = yapx.arg(
            default=False, group="myke args", exclusive=True, flags=["--myke-create"]
        )

    prog: str = _file if _file else MYKE_VAR_NAME

    parser = yapx.ArgumentParser(
        prog=prog,
        add_help=False,
        description="Parameters for Myke:",
    )
    parser.add_arguments(MykeArgs)

    myke_args: MykeArgs
    task_args: List[str]
    myke_args, task_args = parser.parse_known_args_to_model(
        sys.argv[1:], args_model=MykeArgs, use_pydantic=False
    )
    assert isinstance(myke_args, MykeArgs)

    if myke_args.version:
        print.text(__version__)
        parser.exit()
    elif myke_args.create:
        write.mykefile(myke_args.file)
        print.text(f"Created: {myke_args.file}")
        parser.exit()

    if myke_args.env_file:
        os.environ.update(read.envfile(myke_args.env_file))

    if _file:
        if not os.path.exists(_file):
            raise FileNotFoundError(_file)
        if not TASKS:
            raise NoTasksFoundError(_file)
        _file = os.path.abspath(_file)

    mykefiles: List[str] = (
        [myke_args.file]
        if os.path.dirname(myke_args.file)
        else [
            y
            for x in myke_args.file_paths
            for y in [os.path.join(x, myke_args.file)]
            if os.path.exists(y)
        ]
    )

    for f in mykefiles:
        f = os.path.abspath(f)
        try:
            if f != _file:
                import_module(f)
            elif not TASKS:
                raise NoTasksFoundError(f)
        except FileNotFoundError:
            parser.print_help()
            print.text(
                f"{os.linesep}"
                f"'{myke_args.file}' not found. Create it using:"
                f"{os.linesep}"
                f"> {prog} --myke-create --myke-file '{f}'"
                f"{os.linesep}"
            )
            parser.exit()

    root_task: Optional[Callable[..., Any]] = TASKS.pop(ROOT_TASK_KEY, None)

    if myke_args.help or (not task_args and not myke_args.task_help_all):
        parser.print_help()
        print.text()
        print.table(
            [{"Command": k, "Source": v.__module__} for k, v in sorted(TASKS.items())],
            tablefmt="github",
        )
        print.text()
        print.text("To view task parameters, see:")
        print.text(f"> {prog} <task-name> --help")
        print.text()
        parser.exit()

    if myke_args.task_help:
        task_args.append("--help")

    yapx.run(
        root_task,
        _args=task_args,
        _prog=prog,
        _print_help=myke_args.task_help_all,
        _use_pydantic=(not myke_args.no_pydantic),
        **TASKS,
    )



from typing import Callable, Dict

try:
    from typing_extensions import Protocol
except ImportError:
    from typing import Protocol


class MykeType(Protocol):
    TASKS: Dict[str, Callable[..., None]]



__all__ = [
    "convert_to_command_string",
    "make_executable",
    "is_version",
]

import os
import re
import stat
from functools import wraps
from importlib.machinery import SourceFileLoader
from typing import Any, List, Optional

from yapx.utils import convert_to_command_string


def split_and_trim_text(txt: Optional[str]) -> List[str]:
    if txt is None:
        return []

    txt = txt.strip()

    if not txt:
        return []

    return [x_trim for x in txt.splitlines() for x_trim in [x.strip()] if x_trim]


def make_executable(file: str) -> None:
    st = os.stat(file)
    os.chmod(file, st.st_mode | stat.S_IEXEC)


def is_version(txt: str) -> bool:
    # https://github.com/pypa/packaging/blob/main/packaging/version.py
    VERSION_PATTERN = r"""
    v?
    (?:
        (?:(?P<epoch>[0-9]+)!)?                           # epoch
        (?P<release>[0-9]+(?:\.[0-9]+)*)                  # release segment
        (?P<pre>                                          # pre-release
            [-_\.]?
            (?P<pre_l>(a|b|c|rc|alpha|beta|pre|preview))
            [-_\.]?
            (?P<pre_n>[0-9]+)?
        )?
        (?P<post>                                         # post release
            (?:-(?P<post_n1>[0-9]+))
            |
            (?:
                [-_\.]?
                (?P<post_l>post|rev|r)
                [-_\.]?
                (?P<post_n2>[0-9]+)?
            )
        )?
        (?P<dev>                                          # dev release
            [-_\.]?
            (?P<dev_l>dev)
            [-_\.]?
            (?P<dev_n>[0-9]+)?
        )?
    )
    (?:\+(?P<local>[a-z0-9]+(?:[-_\.][a-z0-9]+)*))?       # local version
"""
    return (
        re.compile(f"^{VERSION_PATTERN}$", flags=re.VERBOSE | re.IGNORECASE).search(txt)
        is not None
    )


class MykeSourceFileLoader(SourceFileLoader):
    """SourceFileLoader that does not output '__pycache__'"""

    def _cache_bytecode(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()

    def set_data(self, *args: Any, **kwargs: Any) -> None:
        raise NotImplementedError()



__all__ = ["sh", "sh_stdout", "sh_stdout_lines"]

import os
import subprocess
from functools import wraps
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from .exceptions import CalledProcessError
from .utils import split_and_trim_text


def sh(
    args: Union[str, Sequence[str]],
    capture_output: Optional[bool] = False,
    echo: Optional[bool] = True,
    check: Optional[bool] = True,
    cwd: Optional[str] = None,
    env: Optional[Dict[str, str]] = None,
    env_update: Optional[Dict[str, str]] = None,
    timeout: Optional[float] = None,
    **kwargs: Any,
) -> Tuple[Optional[str], Optional[str], int]:

    kwargs["args"] = args
    kwargs["cwd"] = cwd
    kwargs["timeout"] = timeout
    kwargs["check"] = False
    kwargs["shell"] = True
    kwargs["text"] = True
    kwargs["capture_output"] = capture_output

    if env:
        env = env.copy()
    else:
        env = os.environ.copy()

    if env_update:
        env.update(env_update)

    kwargs["env"] = env

    p: subprocess.CompletedProcess = subprocess.run(**kwargs)
    assert isinstance(p, subprocess.CompletedProcess)

    if capture_output and echo:
        if p.stdout:
            print(p.stdout.rstrip(os.linesep))
        if p.stderr:
            print(p.stderr.rstrip(os.linesep))

    if check and p.returncode:
        raise CalledProcessError(p.returncode, p.args, p.stdout, p.stderr)

    return p.stdout, p.stderr, p.returncode


@wraps(sh)
def sh_stdout_lines(*args: Any, **kwargs: Any) -> List[str]:
    kwargs["capture_output"] = True
    kwargs["echo"] = kwargs.get("echo", False)
    stdout, *_ = sh(*args, **kwargs)
    return split_and_trim_text(stdout)


@wraps(sh)
def sh_stdout(*args: Any, **kwargs: Any) -> str:
    return os.linesep.join(sh_stdout_lines(*args, **kwargs))



__all__ = ["text", "lines", "json", "yaml", "cfg", "dotfile", "envfile"]

# import urllib.request
from functools import partial, wraps
from typing import Any, Callable, Dict, List

try:
    from typing_extensions import TypeGuard
except ImportError:
    from typing import TypeGuard


def _is_simple_dict(candidate: Any) -> TypeGuard[Dict[str, Any]]:
    return isinstance(candidate, dict) and all(isinstance(k, str) for k in candidate)


def _read_simple_dict(reader: Callable[..., Any], **kwargs: Any) -> Dict[str, Any]:
    content: Any = reader(**kwargs)
    if not _is_simple_dict(content):
        raise TypeError("expected a dictionary with string keys")
    return content


def text(path: str, encoding: str = "utf-8") -> str:
    with open(path, "r", encoding=encoding) as f:
        return f.read().strip()


@wraps(text)
def lines(*args: str, **kwargs: str) -> List[str]:
    return [y for x in text(*args, **kwargs).splitlines() for y in [x.strip()] if y]


@wraps(text)
def json(*args: str, **kwargs: str) -> Dict[str, Any]:
    import json as _json

    return _read_simple_dict(partial(_json.loads, text(*args, **kwargs)))


@wraps(text)
def yaml(*args: str, **kwargs: str) -> Dict[str, Any]:
    import yaml as _yaml

    return _read_simple_dict(partial(_yaml.safe_load, text(*args, **kwargs)))


@wraps(text)
def yaml_all(*args: str, **kwargs: str) -> List[Dict[str, Any]]:
    import yaml as _yaml

    def _yaml_all(txt: str) -> List[Dict[str, Any]]:
        return [_read_simple_dict(lambda y: y, y=x) for x in _yaml.safe_load_all(txt)]

    return _yaml_all(text(*args, **kwargs))


@wraps(text)
def toml(*args: str, **kwargs: str) -> Dict[str, Any]:
    try:
        import tomli as _toml
    except ImportError:
        import tomllib as _toml

    return _read_simple_dict(partial(_toml.loads, text(*args, **kwargs)))


@wraps(text)
def cfg(*args: str, **kwargs: str) -> Dict[str, Any]:
    from configparser import ConfigParser

    def _read_cfg(txt: str) -> Dict[str, Any]:
        cp = ConfigParser()
        cp.read_string(txt)
        return {x: dict(cp.items(x)) for x in cp.sections()}

    return _read_simple_dict(partial(_read_cfg, text(*args, **kwargs)))


@wraps(cfg)
def ini(*args: str, **kwargs: str) -> Dict[str, Any]:
    return cfg(*args, **kwargs)


@wraps(text)
def dotfile(*args: str, **kwargs: str) -> Dict[str, str]:
    from io import StringIO

    from dotenv import dotenv_values

    return _read_simple_dict(
        partial(dotenv_values, stream=StringIO(text(*args, **kwargs)))
    )


@wraps(dotfile)
def envfile(*args: str, **kwargs: str) -> Dict[str, str]:
    return dotfile(*args, **kwargs)


def _url(addr: str, **kwargs: Any) -> Any:
    import requests

    addr = kwargs.pop("url", addr)
    method: str = kwargs.pop("method", "GET")
    timeout: float = kwargs.pop("timeout", 10)
    resp: requests.Response = requests.request(
        method=method, url=addr, timeout=timeout, **kwargs
    )
    return resp

    # req = urllib.request.Request(url=addr, **kwargs)
    # with urllib.request.urlopen(req) as resp:
    #     return str(resp.read())


def url(addr: str, **kwargs: Any) -> str:
    resp_text: str = _url(addr=addr, **kwargs).text
    return resp_text


def url_dict(addr: str, **kwargs: Any) -> Dict[str, Any]:
    resp: Any = _url(addr=addr, **kwargs).json()
    resp_dict: Dict[str, Any] = _read_simple_dict(lambda: resp)
    return resp_dict
